//
// A Multimedia Traffic Simulation
//
// 
// (c) Copyright ensi, 1997.
//
// packet generation class - source file
//

#include "packet_gen.h"


c_packet_generator::c_packet_generator(int lower_bound, int upper_bound, int p_size, int s_throughput)
{
  image_seq = 0;
  packet_seq = 0;
  current_image_size_in_packets = 1;
  last_packet_id = 0;
  isize_lower_bound = lower_bound;
  isize_upper_bound = upper_bound;
  packet_size = p_size;
  source_throughput = s_throughput;
  srand(1);
}


c_packet_generator::c_packet_generator()
{
  image_seq = 0;
  packet_seq = 0;
  current_image_size_in_packets = 1;
  last_packet_id = 0;
  isize_lower_bound = DEFAULT_LOWER_BOUND;
  isize_upper_bound = DEFAULT_UPPER_BOUND;
  packet_size = DEFAULT_PACKET_SIZE;
  source_throughput = DEFAULT_SOURCE_THROUGHPUT;
  srand(1);
}


int c_packet_generator::gen_image_size()
{
     // generates a number between isize_lower_bound
     // and isize_upper_bound
   return (int)(isize_lower_bound +
               (isize_upper_bound - isize_lower_bound)*(rand()/(1.0*RAND_MAX)));
}


int c_packet_generator::gen_packet(c_packet *pack, double c_date)
{
  ASSERT_VALID(pack);
  
     // is this the first packet, or a part of an image
     // being processed ?
  if (last_packet_id == current_image_size_in_packets - 1)
    {
      current_image_size_in_packets = (int) (gen_image_size()/(1.0*packet_size)) + 1;
      image_seq++;
      packet_seq=0;
    }
  else
    {
      packet_seq++;
    }

  last_packet_id = packet_seq;
  
  pack->set_image_id(image_seq);
  pack->set_packet_id(packet_seq);
  pack->set_nb_packets(current_image_size_in_packets);
  pack->set_gen_date(c_date);
  
  return 0;
}


int c_packet_generator::get_source_throughput()
{
  return source_throughput;
}


int c_packet_generator::get_packet_size()
{
  return packet_size;
}


int c_packet_generator::get_image_count()
{
  return image_seq;
}

c_packet_generator::~c_packet_generator()
{
}
