//
// A Multimedia Traffic Simulation
//
// 
// (c) Copyright ensi, 1997.
//
// router class - source file
//

#include "router.h"

c_router::c_router()
{
  current_date = 0.0;
  sim_duration = DEFAULT_SIM_DURATION;
  last_departure = 0.0;
  net_capacity = DEFAULT_NET_CAPACITY;
  queue_capacity = DEFAULT_QUEUE_CAPACITY;
  time_line = new c_time_line;
  queue_size = 0;
}


c_router::c_router(int net_cap, double sd, int queue_cap)
{
  current_date = 0.0;
  sim_duration = sd;
  last_departure = 0.0;
  net_capacity = net_cap;
  queue_capacity = queue_cap;
  time_line = new c_time_line;
  queue_size = 0;
}


int c_router::get_net_capacity()
{
  return net_capacity;
}


int c_router::init_simulation()
{
  c_event first_event;
  
  packet_gen->gen_packet(first_event.packet, current_date);
  cout << "First event generated" << endl;                    //-
  
  time_line->insert_event(EVENT_ARRIVAL, current_date, first_event.packet);
  cout << "First event inserted" << endl;                    //-
  
  return OK;
}


int c_router::start_simulation()
{
  return OK;
}


c_router::~c_router()
{
}


c_router_fifo::c_router_fifo() : c_router()
{
  packet_gen = new c_packet_generator();
}


c_router_fifo::c_router_fifo(int lb, int ub, int st, int p_size, int net_cap, double sd, int qc) : c_router(net_cap, sd, qc)
{
  packet_gen = new c_packet_generator(lb, ub, p_size, st);
}


int c_router_fifo::start_simulation ()
{
  c_event current_event, new_event;
  int rejected_images = 0;
  
  int discard_flag = FALSE, success_flag = TRUE;
  int last_successful_packet = -1;
  int current_image = 1;
  int this_image;
  int current_image_size = 0;

    // time between the generation of two consecutive packets
  double pack_intergen;
  double service_time;

  collector.register_setting("policy (FIFO)", POLICY_FIFO);
  collector.register_setting("packet_size", packet_gen->get_packet_size());
  collector.register_setting("source_throughput", packet_gen->get_source_throughput());
  collector.register_setting("net_capacity", get_net_capacity());

  while (current_date < sim_duration)
    {
      // if there are no more events, exit simulation
      if (time_line->is_empty ())
	return DO_NOT_MIND;

      // pick up the first (earliest) event on the time_line
      // this event is automatically discarded
      current_event = time_line->retrieve_event ();
      current_date = current_event.get_event_date();

      cout << "current date = " << current_date << " qsize = " << queue_size << endl;	//-

      switch (current_event.get_event_type ())
	{
	case EVENT_ARRIVAL:
	  cout << "start_simulation(): processing EVENT_ARRIVAL" << endl;
	 
	  // generating one arrival
	  if (queue_size >= queue_capacity)
	    {
	      discard_flag = TRUE;
	      if (current_event.packet->is_last()) 
	        {
	          rejected_images++;
	          discard_flag = FALSE;
	          collector.register_success_rate(collector.get_safe_images()/(1.0*(collector.get_safe_images()+rejected_images)), current_date);
	        }
              packet_gen->gen_packet(new_event.packet, current_date);
	      pack_intergen = packet_gen->get_source_throughput()*current_event.packet->get_nb_packets();
              time_line->insert_event(EVENT_ARRIVAL, current_date + 1/pack_intergen, new_event.packet);
	      break;
	    }

          packet_gen->gen_packet(new_event.packet, current_date);
	  pack_intergen = packet_gen->get_source_throughput()*current_event.packet->get_nb_packets(); 
	  time_line->insert_event(EVENT_ARRIVAL, current_date + 1/pack_intergen, new_event.packet);
  	  queue_size++;

          // generating one entry
  	  service_time = 8*packet_gen->get_packet_size()/(1.0*get_net_capacity());
	  if (current_date < last_departure)
	    {
	      time_line->insert_event(EVENT_ENTRY, last_departure, current_event.packet);
	      last_departure += service_time;
	    }
	  else
	    {
	      time_line->insert_event(EVENT_ENTRY, current_date, current_event.packet);
	      last_departure = current_date + service_time;
	    }
        
          if (current_event.packet->is_last() && discard_flag) 
	    {
              rejected_images++;
              discard_flag = FALSE;
	      collector.register_success_rate((int)collector.get_safe_images()/(1.0*(collector.get_safe_images()+rejected_images)), current_date);	          
	    }
          break;

	case EVENT_ENTRY:
  	  cout << "start_simulation(): processing EVENT_ENTRY" << endl;
          service_time = 8*packet_gen->get_packet_size()/(1.0*get_net_capacity());
	  time_line->insert_event(EVENT_DEPARTURE, current_date + service_time, current_event.packet);
	  break;

	case EVENT_DEPARTURE:
	  cout << "start_simulation(): processing EVENT_DEPARTURE >> " << current_event.packet->get_image_id() << "-" << current_event.packet->get_packet_id() << endl;
	  queue_size--;
	  
	  // read the outgoing packet's image id
	  this_image = current_event.packet->get_image_id();
	  
	  // does it belong to a new image ?
	  if ((this_image != current_image) || (current_image == 1))
	    {
	      if ((current_image_size == last_successful_packet + 1) && success_flag)
	        {
	          // gather stats on success rate
                  pack_intergen = 1/(packet_gen->get_source_throughput()*current_event.packet->get_nb_packets());
	          collector.register_rt(current_date - (current_event.packet->get_gen_date() - pack_intergen*(current_event.packet->get_nb_packets() - 1)), current_date, /*current_event.packet->get_image_id()*/ current_image);
	          collector.inc_safe_images();
	          collector.register_success_rate((double)collector.get_safe_images()/(1.0*(collector.get_safe_images()+rejected_images)), current_date);
	        }        
	        
	      // reset reject detection settings
	      success_flag = TRUE;
	      last_successful_packet = -1;
	      current_image = this_image;
	      current_image_size = current_event.packet->get_nb_packets();
	    }
	    
	  // have we lost the expected packet ?
	  // if yes, unset the success flag. That is, declare the image lost
	  if (current_event.packet->get_packet_id() != last_successful_packet + 1)
	    success_flag = FALSE;
          
          // set the expected flag  
	  last_successful_packet = current_event.packet->get_packet_id();
 	  break;
	}
    }
    
  collector.register_setting("safe_images", collector.get_safe_images());
  collector.register_setting("generated_images", packet_gen->get_image_count());

  return OK;
}


c_router_fifo::~c_router_fifo()
{
}


c_router_epd::c_router_epd() : c_router()
{
  epd_threshold = DEFAULT_EPD_THRESHOLD;
}


c_router_epd::c_router_epd(int lb, int ub, int st, int p_size, int tepd, int net_cap, double sd, int qc) : c_router(net_cap, sd, qc)
{
  epd_threshold = tepd;
  packet_gen = new c_packet_generator(lb, ub, p_size, st);
}


int c_router_epd::get_epd_threshold()
{
  return epd_threshold;
}


int c_router_epd::start_simulation ()
{
  c_event current_event, new_event;
  int rejected_images = 0;
  
  int discard_image_flag = FALSE, discard_flag = FALSE, success_flag = TRUE;
  int last_successful_packet = -1;
  int current_image = 1, c_image = 1;
  int this_image, t_image;
  int current_image_size = -1;

    // time between the generation of two consecutive packets
  double pack_intergen;
  double service_time;

  collector.register_setting("policy (EPD)", POLICY_EPD);
  collector.register_setting("packet_size", packet_gen->get_packet_size());
  collector.register_setting("source_throughput", packet_gen->get_source_throughput());
  collector.register_setting("EPD_threshold", epd_threshold);
  collector.register_setting("net_capacity", get_net_capacity());

  while (current_date < sim_duration)
    {
      // if there are no more events, exit simulation
      if (time_line->is_empty ())
        return DO_NOT_MIND;


      // pick up the first (earliest) event on the time_line
      // this event is automatically discarded
      current_event = time_line->retrieve_event ();
      current_date = current_event.get_event_date();

      cout << "current date = " << current_date << " qsize = " << queue_size << "   packet_id = " << current_event.packet->get_packet_id() << endl;	//-

      switch (current_event.get_event_type ())
	{
	case EVENT_ARRIVAL:
	  cout << "start_simulation(): processing EVENT_ARRIVAL" << endl;
	 
	  // generating one arrival
          packet_gen->gen_packet(new_event.packet, current_date);
	  pack_intergen = packet_gen->get_source_throughput()*current_event.packet->get_nb_packets(); 
	  time_line->insert_event(EVENT_ARRIVAL, current_date + 1/pack_intergen, new_event.packet);
	  
	  if (queue_size >= queue_capacity)
	    {
	      cout << "Rejecting ..." << endl;
	      discard_flag = TRUE;
	      if (current_event.packet->is_last())
	        {
	          rejected_images++;
	          discard_flag = FALSE;
	          discard_image_flag = FALSE;
	          collector.register_success_rate(collector.get_safe_images()/(1.0*(collector.get_safe_images()+rejected_images)), current_date);
	        }
	      break;
	    }

	  t_image = current_event.packet->get_image_id();
	  if (queue_size >= epd_threshold)
	    {
              cout << "q: beyond the threshold" << endl; //-
	      if (t_image != c_image)
	        {
	          if (discard_image_flag)
	            {
	              rejected_images++;
                      collector.register_success_rate(collector.get_safe_images()/(1.0*(collector.get_safe_images()+rejected_images)), current_date);
                    }
	          discard_image_flag = TRUE;
  	          c_image = t_image;
	          break;	          
	        }	      
	    }
  	  c_image = t_image;
  	  queue_size++;

          // generating one entry
  	  service_time = 8*packet_gen->get_packet_size()/(1.0*get_net_capacity());
	  if (current_date < last_departure)
	    {
	      time_line->insert_event(EVENT_ENTRY, last_departure, current_event.packet);
	      last_departure += service_time;
	    }
	  else
	    {
	      time_line->insert_event(EVENT_ENTRY, current_date, current_event.packet);
	      last_departure = current_date + service_time;
	    }
        
          if (current_event.packet->is_last() && discard_flag) 
	    {
              rejected_images++;
              discard_flag = FALSE;
              discard_image_flag = FALSE;
	      collector.register_success_rate((int)collector.get_safe_images()/(1.0*(collector.get_safe_images()+rejected_images)), current_date);
	    }
          break;

	case EVENT_ENTRY:
  	  cout << "start_simulation(): processing EVENT_ENTRY" << endl;
          service_time = 8*packet_gen->get_packet_size()/(1.0*get_net_capacity());
	  time_line->insert_event(EVENT_DEPARTURE, current_date + service_time, current_event.packet);
	  break;

	case EVENT_DEPARTURE:
	  cout << "start_simulation(): processing EVENT_DEPARTURE >> " << current_event.packet->get_image_id() << "-" << current_event.packet->get_packet_id() << endl;
	  queue_size--;
	  
	  // read the outgoing packet's image id
	  this_image = current_event.packet->get_image_id();
	  
	  // does it belong to a new image ?
	  if ((this_image != current_image) || (current_image == 1))
	    {
	      if ((current_image_size == last_successful_packet + 1) && success_flag)
	        {
	          // gather stats on success rate
                  pack_intergen = 1/(packet_gen->get_source_throughput()*current_event.packet->get_nb_packets());
	          collector.register_rt(current_date - (current_event.packet->get_gen_date() - pack_intergen*(current_event.packet->get_nb_packets() - 1)), current_date, current_image);
	          collector.inc_safe_images();
	          collector.register_success_rate((double)collector.get_safe_images()/(1.0*(collector.get_safe_images()+rejected_images)), current_date);
	        }        
	        
	      // reset reject detection settings
	      success_flag = TRUE;
	      last_successful_packet = -1;
	      current_image = this_image;
	      current_image_size = current_event.packet->get_nb_packets();
	    }
	    
	  // have we lost the expected packet ?
	  // if yes, unset the success flag. That is, declare the image lost
	  if (current_event.packet->get_packet_id() != last_successful_packet + 1)
	    success_flag = FALSE;
          
	  last_successful_packet = current_event.packet->get_packet_id();
 	  break;
	}
    }
    
  collector.register_setting("safe_images", collector.get_safe_images());
  collector.register_setting("generated_images", packet_gen->get_image_count());

  return OK;
}


c_router_epd::~c_router_epd()
{
}


int main(int argc, char **argv)
{  
  if (argc == 1)
    {
      c_router_fifo router;
      router.init_simulation();
      router.start_simulation();
      return 0;
    }

  int p_policy             = DEFAULT_POLICY;
  int p_source_throughput  = DEFAULT_SOURCE_THROUGHPUT;
  int p_packet_size        = DEFAULT_PACKET_SIZE;
  int p_epd_threshold      = DEFAULT_EPD_THRESHOLD;
  int p_net_capacity       = DEFAULT_NET_CAPACITY;
  int p_queue_capacity     = DEFAULT_QUEUE_CAPACITY;
  int p_sim_duration       = DEFAULT_SIM_DURATION;

  int p_parser = 1;
  while(p_parser < argc)
  {
  switch (argv[p_parser][1])
    {
      case 'p' :
        if (argv[p_parser+1][0]=='E' || argv[p_parser+1][0]=='e')
                   p_policy = POLICY_EPD;
        p_parser++;
        break;
        
      case 's' :
        sscanf(argv[p_parser+1], "%d", &p_source_throughput);
        p_parser++;
        break;
        
      case 'q' :
        sscanf(argv[p_parser+1], "%d", &p_packet_size);
        p_parser++;
        break;
      
      case 't' :
        sscanf(argv[p_parser+1], "%d", &p_epd_threshold);
        p_parser++;
        break;

      case 'n' :
        sscanf(argv[p_parser+1], "%d", &p_net_capacity);
        p_parser++;
        break;

      case 'd' :
        sscanf(argv[p_parser+1], "%d", &p_sim_duration);
        p_parser++;
        break;
        
      case 'c' :
        sscanf(argv[p_parser+1], "%d", &p_queue_capacity);
        p_parser++;
        break;
            
      case '-' :
        cout << "Usage: " << argv[0] << " [-p policy] [-s source_throughput] [-q packet_size] [-t epd_threshold] [-n net_capacity] [-d sim_duration] [-c queue_capacity]" << endl <<
        "p	FIFO or EPD (defaults to FIFO)" << endl <<
        "s	in images per second (defaults to " << DEFAULT_SOURCE_THROUGHPUT << ")" << endl <<
        "q	in bits (defaults to " << DEFAULT_PACKET_SIZE << ")" << endl <<
        "t	in packets (defaults to " << DEFAULT_EPD_THRESHOLD << ")" << endl <<
        "n	in packets (defaults to " << DEFAULT_NET_CAPACITY << ")"<< endl <<
        "d	in bits per second (defaults to " << DEFAULT_SIM_DURATION << ")" << endl <<
        "c	in packets (defaults to " << DEFAULT_QUEUE_CAPACITY << ")" << endl;
          p_parser++;
        return 0;
        
      default:
        p_parser++;
    }
    }

  if (p_policy == POLICY_FIFO)
    {
      c_router_fifo router(0, 1024*10-1, p_source_throughput, p_packet_size, p_net_capacity, p_sim_duration, p_queue_capacity);
      router.init_simulation();
      router.start_simulation();
    }
  else
    {
      c_router_epd router(0, 1024*10-1, p_source_throughput, p_packet_size, p_epd_threshold, p_net_capacity, p_sim_duration, p_queue_capacity);
      router.init_simulation();
      router.start_simulation();
    }
    
  return 0;
}
























